from pathlib import Path

from jinja2 import Environment, FileSystemLoader


def create_jinja_environment(templates_path: Path) -> Environment:
    return Environment(loader=FileSystemLoader(templates_path))



import textwrap
from functools import reduce
from http import HTTPStatus
from typing import Sequence

from spec2sdk.base import Model
from spec2sdk.models.converters import converters
from spec2sdk.models.identifiers import make_variable_name
from spec2sdk.openapi.entities import Endpoint, Parameter, ParameterLocation, Path, RequestBody, Response


def wordwrap(s: str, width: int) -> str:
    return "\n".join(
        textwrap.wrap(
            s,
            width=width,
            expand_tabs=False,
            replace_whitespace=False,
            break_long_words=True,
            break_on_hyphens=True,
        ),
    )


def indent(s: str, width: int) -> str:
    return textwrap.indent(s, " " * width).strip()


class ParameterView:
    def __init__(self, parameter: Parameter):
        self.__parameter = parameter

    @property
    def name(self) -> str:
        return make_variable_name(self.__parameter.name)

    @property
    def original_name(self) -> str:
        return self.__parameter.name

    @property
    def description(self) -> str:
        return self.__parameter.description or ""

    @property
    def type_hint(self) -> str:
        return converters.convert(self.__parameter.data_type).type_hint

    @property
    def required(self) -> bool:
        return self.__parameter.required

    @property
    def default_value(self) -> str:
        return repr(self.__parameter.default_value)


class PathView:
    def __init__(self, path: Path):
        self.__path = path

    @property
    def parameters(self) -> Sequence[ParameterView]:
        return tuple(ParameterView(parameter) for parameter in self.__path.parameters)

    @property
    def path_parameters(self) -> Sequence[ParameterView]:
        return tuple(
            ParameterView(parameter)
            for parameter in self.__path.parameters
            if parameter.location == ParameterLocation.PATH
        )

    @property
    def query_parameters(self) -> Sequence[ParameterView]:
        return tuple(
            ParameterView(parameter)
            for parameter in self.__path.parameters
            if parameter.location == ParameterLocation.QUERY
        )

    @property
    def url(self) -> str:
        return reduce(
            lambda url, parameter: url.replace(f"{{{parameter.original_name}}}", f"{{{parameter.name}}}"),
            self.parameters,
            self.__path.path,
        )


class ResponseView:
    def __init__(self, response: Response):
        self.__response = response

    @property
    def type_hint(self) -> str:
        return (
            converters.convert(self.__response.content.data_type).type_hint if self.__response.content else repr(None)
        )

    @property
    def has_content(self) -> bool:
        return self.__response.content is not None

    @property
    def media_type(self) -> str | None:
        return self.__response.content.media_type if self.__response.content else None

    @property
    def status_code(self) -> HTTPStatus:
        return HTTPStatus(int(self.__response.status_code))


class RequestBodyView:
    def __init__(self, request_body: RequestBody):
        self.__request_body = request_body

    @property
    def name(self) -> str:
        return make_variable_name(converters.convert(self.__request_body.content.data_type).name)

    @property
    def description(self) -> str | None:
        return self.__request_body.description

    @property
    def type_hint(self) -> str:
        return converters.convert(self.__request_body.content.data_type).type_hint

    @property
    def content_type(self) -> str:
        return self.__request_body.content.media_type


class MethodParameter(Model):
    name: str
    description: str | None
    type_hint: str
    required: bool
    default_value: str | None


class EndpointView:
    def __init__(self, endpoint: Endpoint, response: Response):
        self.__endpoint = endpoint
        self.__response = response

    @property
    def method_name(self) -> str:
        name = make_variable_name(self.__endpoint.operation_id)

        if self.response.status_code != HTTPStatus.OK:
            name += f"_expect_{self.__response.status_code}"

        return name

    @property
    def method_parameters(self) -> Sequence[MethodParameter]:
        parameters = []

        if request_body := self.request_body:
            parameters.append(
                MethodParameter(
                    name=request_body.name,
                    description=request_body.description,
                    type_hint=request_body.type_hint,
                    required=True,
                    default_value=None,
                ),
            )

        parameters += [
            MethodParameter(
                name=parameter.name,
                description=parameter.description,
                type_hint=parameter.type_hint,
                required=parameter.required,
                default_value=parameter.default_value,
            )
            for parameter in self.path.parameters
        ]

        return tuple(parameters)

    @property
    def http_method(self) -> str:
        return self.__endpoint.method

    @property
    def path(self) -> PathView:
        return PathView(self.__endpoint.path)

    @property
    def summary(self) -> str | None:
        return self.__endpoint.summary

    @property
    def docstring(self) -> str:
        result = []

        if self.summary:
            result.append(wordwrap(self.summary, width=99))

        if self.method_parameters:
            result.append(
                "\n".join(
                    indent(wordwrap(f":param {parameter.name}: {parameter.description or ''}", width=99), width=7)
                    for parameter in self.method_parameters
                ),
            )

        return "\n\n".join(result)

    @property
    def request_body(self) -> RequestBodyView | None:
        return RequestBodyView(self.__endpoint.request_body) if self.__endpoint.request_body else None

    @property
    def response(self) -> ResponseView:
        return ResponseView(response=self.__response)



from pathlib import Path
from typing import Sequence

from spec2sdk.base import Model
from spec2sdk.templating import create_jinja_environment


class Import(Model):
    name: str
    package: str

    def __hash__(self):
        return hash((self.name, self.package))


def render_imports(imports: Sequence[Import]) -> str:
    return (
        create_jinja_environment(templates_path=Path(__file__).parent / "templates")
        .get_template("imports.j2")
        .render(imports=imports)
    )



from functools import reduce
from typing import Any, ClassVar, Final, Sequence
from urllib.parse import urldefrag, urljoin
from urllib.request import urlopen

import yaml
from pydantic import BaseModel, ConfigDict

from spec2sdk.openapi.exceptions import CircularReference

SCHEMA_NAME_FIELD: Final[str] = "x-schema-name"


class SchemaLoader(BaseModel):
    model_config = ConfigDict(frozen=True)
    _schema_cache: ClassVar[dict] = {}
    schema_url: str

    def _read_schema(self) -> dict:
        if self.schema_url not in self._schema_cache:
            self._schema_cache[self.schema_url] = yaml.safe_load(stream=urlopen(self.schema_url))

        return self._schema_cache[self.schema_url]

    def get_schema_fragment(self, path: str) -> dict:
        return reduce(
            lambda acc, key: acc[key],
            filter(None, path.split("/")),
            self._read_schema(),
        )

    def parse(self, reference: str) -> "SchemaLoader":
        reference_url, _ = urldefrag(reference)
        new_schema_url = urljoin(self.schema_url, reference_url)

        return self if new_schema_url == self.schema_url else SchemaLoader(schema_url=new_schema_url)


class ResolvingParser:
    def __init__(self):
        self._resolved_references = {}

    def _resolve_schema(self, schema_loader: SchemaLoader, schema: dict, parent_reference_ids: Sequence[str]) -> dict:
        def resolve_value(value: Any) -> Any:
            if isinstance(value, dict):
                return self._resolve_schema(schema_loader, value, parent_reference_ids)
            elif isinstance(value, list):
                return [resolve_value(item) for item in value]
            else:
                return value

        def resolve_reference(reference: str) -> dict:
            reference_schema_loader = schema_loader.parse(reference)
            _, reference_fragment_path = urldefrag(reference)
            reference_id = f"{reference_schema_loader.schema_url}#{reference_fragment_path}"

            if reference_id in parent_reference_ids:
                raise CircularReference(f"Circular reference found in {reference}")

            if reference_id not in self._resolved_references:
                self._resolved_references[reference_id] = {
                    SCHEMA_NAME_FIELD: reference_fragment_path.rsplit("/", 1)[-1],
                } | self._resolve_schema(
                    schema_loader=reference_schema_loader,
                    schema=reference_schema_loader.get_schema_fragment(path=reference_fragment_path),
                    parent_reference_ids=(*parent_reference_ids, reference_id),
                )

            return self._resolved_references[reference_id]

        # Object containing $ref cannot be extended except for summary and description fields.
        # Values of these fields can only be overridden if they are defined on the referenced object.
        # https://spec.openapis.org/oas/v3.1.1.html#reference-object
        return (
            # Currently, we don't separate properties defined on the referenced object from those
            # defined on the object containing $ref. That's why we ignore summary and description fields,
            # which can be overridden by the object containing $ref.
            resolve_reference(schema["$ref"])
            if "$ref" in schema
            else {key: resolve_value(value) for key, value in schema.items()}
        )

    def parse(self, schema_url: str) -> dict:
        schema_loader = SchemaLoader(schema_url=schema_url)

        return self._resolve_schema(
            schema_loader=schema_loader,
            schema=schema_loader.get_schema_fragment(path="/"),
            parent_reference_ids=(),
        )



from typing import Any, TypedDict

from spec2sdk.models.entities import (
    BinaryType,
    BooleanType,
    EnumMember,
    EnumType,
    FloatType,
    IntegerType,
    ListType,
    LiteralType,
    ModelField,
    ModelType,
    NoneType,
    StrEnumType,
    StringType,
    UnionType,
)
from spec2sdk.models.identifiers import make_class_name, make_constant_name, make_variable_name
from spec2sdk.models.predicates import is_binary_format, is_enum, is_instance, is_literal, is_str_enum
from spec2sdk.openapi.entities import (
    AllOfDataType,
    AnyOfDataType,
    ArrayDataType,
    BooleanDataType,
    DataType,
    Enumerator,
    IntegerDataType,
    NullDataType,
    NumberDataType,
    ObjectDataType,
    OneOfDataType,
    StringDataType,
)
from spec2sdk.registry import Registry

converters = Registry()


class CommonFields(TypedDict):
    name: str | None
    description: str | None
    default_value: Any


def convert_common_fields(data_type: DataType) -> CommonFields:
    return CommonFields(
        name=make_class_name(data_type.name) if data_type.name else None,
        description=data_type.description,
        default_value=data_type.default_value,
    )


@converters.register(predicate=is_instance(StringDataType))
def convert_string(data_type: StringDataType) -> StringType:
    return StringType(
        **convert_common_fields(data_type),
        pattern=data_type.pattern,
        min_length=data_type.min_length,
        max_length=data_type.max_length,
    )


@converters.register(predicate=is_instance(IntegerDataType))
def convert_integer(data_type: IntegerDataType) -> IntegerType:
    return IntegerType(
        **convert_common_fields(data_type),
        minimum=data_type.minimum,
        maximum=data_type.maximum,
        exclusive_minimum=data_type.exclusive_minimum,
        exclusive_maximum=data_type.exclusive_maximum,
        multiple_of=data_type.multiple_of,
    )


@converters.register(predicate=is_instance(NumberDataType))
def convert_number(data_type: NumberDataType) -> FloatType:
    return FloatType(
        **convert_common_fields(data_type),
        minimum=data_type.minimum,
        maximum=data_type.maximum,
        exclusive_minimum=data_type.exclusive_minimum,
        exclusive_maximum=data_type.exclusive_maximum,
        multiple_of=data_type.multiple_of,
    )


@converters.register(predicate=is_instance(BooleanDataType))
def convert_boolean(data_type: BooleanDataType) -> BooleanType:
    return BooleanType(**convert_common_fields(data_type))


@converters.register(predicate=is_instance(ObjectDataType))
def convert_object(data_type: ObjectDataType) -> ModelType:
    return ModelType(
        **convert_common_fields(data_type),
        base_models=(),
        fields=tuple(
            ModelField(
                name=make_variable_name(prop.name),
                alias=prop.name,
                type_hint=inner_py_type.type_hint,
                description=inner_py_type.description if inner_py_type.name is None else None,
                default_value=inner_py_type.default_value,
                is_required=prop.is_required,
                inner_py_type=inner_py_type,
            )
            for prop in data_type.properties
            if (
                inner_py_type := converters.convert(
                    prop.data_type
                    if prop.is_required
                    else OneOfDataType(
                        name=None,
                        description=None,
                        default_value=None,
                        enumerators=None,
                        data_types=(
                            prop.data_type,
                            NullDataType(
                                name=None,
                                description=None,
                                default_value=None,
                                enumerators=None,
                            ),
                        ),
                    ),
                )
            )
        ),
        arbitrary_fields_allowed=data_type.additional_properties,
    )


@converters.register(predicate=is_instance(ArrayDataType))
def convert_array(data_type: ArrayDataType) -> ListType:
    return ListType(
        **convert_common_fields(data_type),
        inner_py_type=converters.convert(data_type.item_type),
        min_items=data_type.min_items,
        max_items=data_type.max_items,
    )


@converters.register(predicate=is_instance((AnyOfDataType, OneOfDataType)))
def convert_one_of(data_type: OneOfDataType | AnyOfDataType) -> UnionType:
    inner_py_types = tuple(map(converters.convert, data_type.data_types))

    return UnionType(
        **convert_common_fields(data_type),
        inner_py_types=inner_py_types,
    )


@converters.register(predicate=is_instance(AllOfDataType))
def convert_all_of(data_type: AllOfDataType) -> ModelType:
    if not all(
        isinstance(inner_data_type, (ObjectDataType, AllOfDataType)) for inner_data_type in data_type.data_types
    ):
        raise TypeError("Non-object data types in allOf are not supported")

    nameless_inner_data_types = tuple(
        inner_data_type for inner_data_type in data_type.data_types if inner_data_type.name is None
    )
    if len(nameless_inner_data_types) > 1:
        raise TypeError("Multiple data types without a name in allOf is not supported")

    model_type = convert_object(
        nameless_inner_data_types[0] if nameless_inner_data_types else data_type.data_types[0],
    )

    return ModelType(
        name=make_class_name(data_type.name),
        description=data_type.description,
        default_value=model_type.default_value,
        base_models=tuple(
            converters.convert(inner_data_type)
            for inner_data_type in data_type.data_types
            if inner_data_type.name is not None
        ),
        fields=model_type.fields if nameless_inner_data_types else (),
        arbitrary_fields_allowed=False,
    )


@converters.register(predicate=is_enum)
def convert_enum(data_type: DataType) -> EnumType:
    def generate_enum_member_name(enumerator: Enumerator) -> str:
        if enumerator.name:
            name = enumerator.name
        elif isinstance(enumerator.value, str):
            name = enumerator.value
        else:
            name = f"{data_type.name}_{enumerator.value}"

        return make_constant_name(name)

    members = tuple(
        EnumMember(name=generate_enum_member_name(member), value=member.value) for member in data_type.enumerators
    )

    default_value = None
    for member in members:
        if member.value == data_type.default_value:
            default_value = member

    return EnumType(
        name=make_class_name(data_type.name),
        description=data_type.description,
        default_value=default_value,
        members=members,
    )


@converters.register(predicate=is_str_enum)
def convert_str_enum(data_type: StringDataType) -> StrEnumType:
    return StrEnumType(**convert_enum(data_type).model_dump())


@converters.register(predicate=is_binary_format)
def convert_binary(data_type: StringDataType) -> BinaryType:
    return BinaryType(**convert_common_fields(data_type))


@converters.register(predicate=is_literal)
def convert_literal(data_type: DataType) -> LiteralType:
    return LiteralType(
        name=None,
        description=data_type.description,
        default_value=data_type.default_value,
        literals=tuple(enumerator.value for enumerator in data_type.enumerators or ()),
    )


@converters.register(predicate=is_instance(NullDataType))
def convert_none(data_type: NullDataType) -> NoneType:
    return NoneType(
        name=data_type.name,
        description=data_type.description,
        default_value=data_type.default_value,
    )



import re
from keyword import iskeyword
from typing import Final, Pattern

from humps import decamelize, pascalize

INVALID_CHARACTERS_PATTERN: Final[Pattern] = re.compile(r"[^0-9a-z]+|^[^a-z]+", flags=re.IGNORECASE)


def make_identifier(name: str) -> str:
    """
    Makes valid Python identifier from the string by removing invalid leading characters
    and replacing invalid characters with underscore.
    """

    name = "_".join(name_part for name_part in INVALID_CHARACTERS_PATTERN.split(name) if name_part)

    # Add underscore to the name if it's a valid Python keyword
    if iskeyword(name):
        name += "_"

    return name


def make_class_name(name: str) -> str:
    return pascalize(make_identifier(name))


def make_constant_name(name: str) -> str:
    return make_identifier(name).upper()


def make_variable_name(name: str) -> str:
    return decamelize(make_identifier((name)))



from enum import StrEnum, auto
from types import NoneType
from typing import Any, Sequence

from spec2sdk.base import Model


class Enumerator[T](Model):
    name: str | None
    value: T


class DataType[T](Model):
    name: str | None
    description: str | None
    default_value: T | None
    enumerators: Sequence[Enumerator[T]] | None


class NumericDataType[T](DataType[T]):
    format: str | None

    # https://json-schema.org/draft/2020-12/json-schema-validation#name-validation-keywords-for-num
    minimum: T | None
    maximum: T | None
    exclusive_minimum: T | None
    exclusive_maximum: T | None
    multiple_of: T | None


class IntegerDataType(NumericDataType[int]):
    pass


class NumberDataType(NumericDataType[float]):
    pass


class StringDataType(DataType[str]):
    format: str | None

    # https://json-schema.org/draft/2020-12/json-schema-validation#name-validation-keywords-for-str
    pattern: str | None
    min_length: int | None
    max_length: int | None


class BooleanDataType(DataType[bool]):
    pass


class NullDataType(DataType[NoneType]):
    pass


class ObjectProperty(Model):
    data_type: DataType
    name: str
    is_required: bool


class ObjectDataType(DataType):
    properties: Sequence[ObjectProperty]
    additional_properties: bool


class ArrayDataType(DataType):
    item_type: DataType

    # https://json-schema.org/draft/2020-12/json-schema-validation#name-validation-keywords-for-arr
    min_items: int | None
    max_items: int | None


class MultiDataType(DataType):
    data_types: Sequence[DataType]


class OneOfDataType(MultiDataType):
    pass


class AnyOfDataType(MultiDataType):
    pass


class AllOfDataType(MultiDataType):
    pass


class ParameterLocation(StrEnum):
    QUERY = auto()
    HEADER = auto()
    PATH = auto()
    COOKIE = auto()


class Parameter(Model):
    name: str
    location: ParameterLocation
    description: str | None
    required: bool
    data_type: DataType
    default_value: Any | None


class Path(Model):
    path: str
    parameters: Sequence[Parameter]


class Content(Model):
    media_type: str
    data_type: DataType


class RequestBody(Model):
    description: str | None
    required: bool
    content: Content


class Response(Model):
    status_code: str
    description: str
    content: Content | None


class Endpoint(Model):
    path: Path
    method: str
    operation_id: str
    summary: str | None
    request_body: RequestBody | None
    responses: Sequence[Response]


class Specification(Model):
    endpoints: Sequence[Endpoint]



from typing import Callable, Type

from spec2sdk.openapi.entities import DataType, StringDataType


def is_instance[T](data_class: Type[T] | tuple[Type[T], ...]) -> Callable[[T], bool]:
    def compare(data_type: T) -> bool:
        return isinstance(data_type, data_class)

    return compare


def is_enum(data_type: DataType) -> bool:
    return data_type.name is not None and data_type.enumerators is not None


def is_str_enum(data_type: DataType) -> bool:
    return isinstance(data_type, StringDataType) and is_enum(data_type)


def is_binary_format(data_type: DataType) -> bool:
    return isinstance(data_type, StringDataType) and data_type.format == "binary"


def is_literal(data_type: DataType) -> bool:
    return data_type.name is None and data_type.enumerators is not None



import textwrap
from abc import abstractmethod
from pathlib import Path
from typing import Any, Sequence

from spec2sdk.base import Model
from spec2sdk.models.imports import Import
from spec2sdk.templating import create_jinja_environment


class PythonType(Model):
    name: str | None
    description: str | None
    default_value: Any

    @property
    @abstractmethod
    def type_hint(self) -> str: ...

    @property
    @abstractmethod
    def imports(self) -> Sequence[Import]: ...

    @abstractmethod
    def render(self) -> str:
        """
        Returns rendered Python type. Method will only be called if type has a name.
        """
        ...

    @property
    def dependency_types(self) -> Sequence["PythonType"]:
        return ()


class LiteralType(PythonType):
    literals: Sequence[Any]

    @property
    def type_hint(self) -> str:
        return "Literal[" + ",".join(repr(literal) for literal in self.literals) + "]"

    @property
    def imports(self) -> Sequence[Import]:
        return (Import(name="Literal", package="typing"),)

    def render(self) -> str:
        return f"type {self.name} = {self.type_hint}"


class EnumMember(Model):
    name: str
    value: Any


class EnumMemberView(Model):
    name: str
    value: str


class EnumType(PythonType):
    members: Sequence[EnumMember]
    default_value: EnumMember | None

    @property
    def type_hint(self) -> str:
        return self.name

    @property
    def imports(self) -> Sequence[Import]:
        return (Import(name="Enum", package="enum"),)

    def render(self) -> str:
        return (
            create_jinja_environment(templates_path=Path(__file__).parent / "templates")
            .get_template("enum.j2")
            .render(
                enum_type=self,
                base_class_name="Enum",
                members=tuple(EnumMemberView(name=member.name, value=member.value) for member in self.members),
            )
        )


class StrEnumType(EnumType):
    @property
    def imports(self) -> Sequence[Import]:
        return (Import(name="StrEnum", package="enum"),)

    def render(self) -> str:
        return (
            create_jinja_environment(templates_path=Path(__file__).parent / "templates")
            .get_template("enum.j2")
            .render(
                enum_type=self,
                base_class_name="StrEnum",
                members=tuple(EnumMemberView(name=member.name, value=f'"{member.value}"') for member in self.members),
            )
        )


class NumericType[T](PythonType):
    default_value: T | None
    minimum: T | None
    maximum: T | None
    exclusive_minimum: T | None
    exclusive_maximum: T | None
    multiple_of: T | None

    @property
    @abstractmethod
    def type_name(self) -> str: ...

    @property
    def _is_constrained_type(self) -> bool:
        return (
            (self.minimum is not None)
            or (self.maximum is not None)
            or (self.exclusive_minimum is not None)
            or (self.exclusive_maximum is not None)
            or (self.multiple_of is not None)
        )

    @property
    def _type_annotation(self) -> str:
        if self._is_constrained_type:
            constraints = ",".join(
                (
                    *((f"ge={self.minimum}",) if self.minimum is not None else ()),
                    *((f"le={self.maximum}",) if self.maximum is not None else ()),
                    *((f"gt={self.exclusive_minimum}",) if self.exclusive_minimum is not None else ()),
                    *((f"lt={self.exclusive_maximum}",) if self.exclusive_maximum is not None else ()),
                    *((f"multiple_of={self.multiple_of}",) if self.multiple_of is not None else ()),
                ),
            )
            return f"Annotated[{self.type_name}, Field({constraints})]"
        else:
            return self.type_name

    @property
    def type_hint(self) -> str:
        return self.name or self._type_annotation

    @property
    def imports(self) -> Sequence[Import]:
        return (
            (
                Import(name="Annotated", package="typing"),
                Import(name="Field", package="pydantic"),
            )
            if self._is_constrained_type
            else ()
        )

    def render(self) -> str:
        return f"type {self.name} = {self._type_annotation}"


class IntegerType(NumericType[int]):
    @property
    def type_name(self) -> str:
        return "int"


class FloatType(NumericType[float]):
    @property
    def type_name(self) -> str:
        return "float"


class BooleanType(PythonType):
    default_value: bool | None

    @property
    def type_hint(self) -> str:
        return self.name or "bool"

    @property
    def imports(self) -> Sequence[Import]:
        return ()

    def render(self) -> str:
        return f"type {self.name} = bool"


class StringType(PythonType):
    default_value: str | None
    pattern: str | None
    min_length: int | None
    max_length: int | None

    @property
    def _is_constrained_type(self) -> bool:
        return bool(self.pattern) or (self.min_length is not None) or (self.max_length is not None)

    @property
    def _type_annotation(self) -> str:
        if self._is_constrained_type:
            constraints = ",".join(
                (
                    *((f'pattern=r"{self.pattern}"',) if self.pattern else ()),
                    *((f"min_length={self.min_length}",) if self.min_length is not None else ()),
                    *((f"max_length={self.max_length}",) if self.max_length is not None else ()),
                ),
            )
            return f"Annotated[str, StringConstraints({constraints})]"
        else:
            return "str"

    @property
    def type_hint(self) -> str:
        return self.name or self._type_annotation

    @property
    def imports(self) -> Sequence[Import]:
        return (
            (
                Import(name="Annotated", package="typing"),
                Import(name="StringConstraints", package="pydantic"),
            )
            if self._is_constrained_type
            else ()
        )

    def render(self) -> str:
        return f"type {self.name} = {self._type_annotation}"


class BinaryType(PythonType):
    @property
    def type_hint(self) -> str:
        return self.name or "bytes"

    @property
    def imports(self) -> Sequence[Import]:
        return ()

    def render(self) -> str:
        return f"type {self.name} = bytes"


class ModelField(Model):
    name: str
    alias: str
    type_hint: str
    description: str | None
    default_value: Any
    is_required: bool
    inner_py_type: PythonType


class ModelFieldView(Model):
    name: str
    type_definition: str


class ModelType(PythonType):
    base_models: Sequence["ModelType"]
    fields: Sequence[ModelField]
    arbitrary_fields_allowed: bool

    @property
    def dependency_types(self) -> Sequence[PythonType]:
        return *tuple(field.inner_py_type for field in self.fields), *self.base_models

    @property
    def type_hint(self) -> str:
        return self.name

    @property
    def imports(self) -> Sequence[Import]:
        return (
            *((Import(name="Field", package="pydantic"),) if len(self.fields) > 0 else ()),
            *((Import(name="ConfigDict", package="pydantic"),) if self.arbitrary_fields_allowed else ()),
        )

    def render(self) -> str:
        def split_long_lines(s: str) -> str:
            return '"' + ' ""'.join(line.replace('"', r"\"") for line in textwrap.wrap(s, width=80)) + '"'

        def create_model_field_view(field: ModelField) -> ModelFieldView:
            attrs = []

            if field.default_value is not None or not field.is_required:
                attrs.append(f"default={repr(field.default_value)}")

            if field.name != field.alias:
                attrs.append(f'alias="{field.alias}"')

            if field.description:
                attrs.append(f"description={split_long_lines(field.description)}")

            return ModelFieldView(
                name=field.name,
                type_definition=field.type_hint + (f" = Field({','.join(attrs)})" if attrs else ""),
            )

        base_class_names = tuple(base_model.name for base_model in self.base_models if base_model.name)

        return (
            create_jinja_environment(templates_path=Path(__file__).parent / "templates")
            .get_template("model.j2")
            .render(
                base_class_name=", ".join(base_class_names) if base_class_names else "Model",
                model_type=self,
                fields=tuple(map(create_model_field_view, self.fields)),
                arbitrary_fields_allowed=self.arbitrary_fields_allowed,
            )
        )


class NoneType(PythonType):
    @property
    def dependency_types(self) -> Sequence[PythonType]:
        return ()

    @property
    def type_hint(self) -> str:
        return self.name or "None"

    @property
    def imports(self) -> Sequence[Import]:
        return ()

    def render(self) -> str:
        return f"type {self.name} = None"


class ListType(PythonType):
    inner_py_type: PythonType
    min_items: int | None
    max_items: int | None

    @property
    def _is_constrained_type(self) -> bool:
        return (self.min_items is not None) or (self.max_items is not None)

    @property
    def _type_annotation(self) -> str:
        type_name = f"list[{self.inner_py_type.type_hint}]"

        if self._is_constrained_type:
            constraints = ",".join(
                (
                    *((f"min_length={self.min_items}",) if self.min_items is not None else ()),
                    *((f"max_length={self.max_items}",) if self.max_items is not None else ()),
                ),
            )
            return f"Annotated[{type_name}, Field({constraints})]"
        else:
            return type_name

    @property
    def dependency_types(self) -> Sequence[PythonType]:
        return (self.inner_py_type,)

    @property
    def type_hint(self) -> str:
        return self.name or self._type_annotation

    @property
    def imports(self) -> Sequence[Import]:
        return (
            (
                Import(name="Annotated", package="typing"),
                Import(name="Field", package="pydantic"),
            )
            if self._is_constrained_type
            else ()
        )

    def render(self) -> str:
        return f"type {self.name} = {self._type_annotation}"


class UnionType(PythonType):
    inner_py_types: Sequence[PythonType]

    @property
    def dependency_types(self) -> Sequence[PythonType]:
        return self.inner_py_types

    @property
    def type_hint(self) -> str:
        return self.name or " | ".join(py_type.type_hint for py_type in self.inner_py_types)

    @property
    def imports(self) -> Sequence[Import]:
        return ()

    def render(self) -> str:
        root_type = " | ".join(py_type.type_hint for py_type in self.inner_py_types)
        return f"type {self.name} = {root_type}"



from pathlib import Path
from typing import Sequence

from spec2sdk.client.views import EndpointView
from spec2sdk.models.converters import converters
from spec2sdk.models.entities import PythonType
from spec2sdk.models.imports import Import, render_imports
from spec2sdk.openapi.entities import Specification
from spec2sdk.openapi.parsers import get_root_data_types
from spec2sdk.templating import create_jinja_environment


def get_imports(py_types: Sequence[PythonType], models_import: Import) -> Sequence[Import]:
    """
    Returns imports for each type in the following cases:
    - if type has a name, it can be imported directly from the models file
      (e.g. `from models import User`)
    - if type doesn't have a name, import types used in the type definition
      (e.g. `from typing import Literal` for `Literal["json", "xml"]`)
    - if type doesn't have a name, import types for inner types
      (e.g. `from models import Product` for `list[Product]`)
    """

    def get_importable_type_names(py_type: PythonType) -> Sequence[str]:
        return (
            (py_type.name,)
            if py_type.name
            else tuple(
                type_name
                for dependency_type in py_type.dependency_types
                for type_name in get_importable_type_names(dependency_type)
            )
        )

    return tuple(
        {
            models_import.model_copy(update={"name": name})
            for py_type in py_types
            for name in get_importable_type_names(py_type)
        }
        | {import_ for py_type in py_types if not py_type.name for import_ in py_type.imports},
    )


def generate_client(spec: Specification, models_import: Import, output_dir: Path):
    root_data_types = get_root_data_types(spec)
    root_python_types = tuple(map(converters.convert, root_data_types))
    imports = get_imports(py_types=root_python_types, models_import=models_import)
    jinja_environment = create_jinja_environment(templates_path=Path(__file__).parent / "templates")

    api_content = jinja_environment.get_template("api.j2").render(
        imports=render_imports(imports),
        endpoints=tuple(
            EndpointView(endpoint=endpoint, response=response)
            for endpoint in spec.endpoints
            for response in endpoint.responses
        ),
    )
    output_dir.joinpath("api.py").write_text(api_content)

    for filename in ("http_client", "api_client"):
        output_dir.joinpath(f"{filename}.py").write_text(
            jinja_environment.get_template(f"{filename}.j2").render(),
        )



from typing import Any, Callable, TypeVar

from pydantic import BaseModel

InputType = TypeVar("InputType")
OutputType = TypeVar("OutputType")

PredicateType = Callable[[InputType], bool]
ConverterType = Callable[[InputType], OutputType]


class AlreadyRegistered(Exception):
    pass


class ConverterNotFound(Exception):
    pass


class ConditionalConverter(BaseModel):
    predicate: PredicateType
    convert: ConverterType


class Registry:
    def __init__(self):
        self.converters: list[ConditionalConverter] = []

    def register(self, predicate: PredicateType):
        def inner(func: ConverterType):
            if func in {converter.convert for converter in self.converters}:
                raise AlreadyRegistered(f"The function '{func.__name__}' is already registered")

            self.converters.append(ConditionalConverter(predicate=predicate, convert=func))

            return func

        return inner

    def convert(self, data) -> Any:
        for converter in reversed(self.converters):
            if converter.predicate(data):
                return converter.convert(data)

        raise ConverterNotFound(f"Converter not found for '{data}'")



from collections import Counter
from graphlib import TopologicalSorter
from pathlib import Path
from typing import Sequence

from spec2sdk.models.converters import converters
from spec2sdk.models.entities import PythonType
from spec2sdk.models.imports import render_imports
from spec2sdk.openapi.entities import Specification
from spec2sdk.openapi.parsers import get_root_data_types
from spec2sdk.templating import create_jinja_environment


def unwrap_py_types(py_types: Sequence[PythonType]) -> Sequence[PythonType]:
    """
    Recursively unwraps all python types and returns all unique python types.
    """
    return tuple(
        {
            dependent_type
            for py_type in py_types
            for dependent_type in (py_type, *unwrap_py_types(py_type.dependency_types))
        },
    )


def get_duplicate_type_names(py_types: Sequence[PythonType]) -> Sequence[str]:
    type_names = tuple(py_type.name for py_type in py_types if py_type.name is not None)
    duplicate_type_names = Counter(type_names) - Counter(set(type_names))
    return tuple(duplicate_type_names.keys())


def generate_models(spec: Specification, output_dir: Path) -> Path:
    root_data_types = get_root_data_types(spec)
    root_py_types = tuple(map(converters.convert, root_data_types))
    all_py_types = unwrap_py_types(root_py_types)

    if duplicate_type_names := get_duplicate_type_names(all_py_types):
        raise Exception(f"Found duplicate type names: {', '.join(duplicate_type_names)}")

    # Types must be sorted in the order of defining their dependencies
    topological_sorter = TopologicalSorter({py_type: py_type.dependency_types for py_type in all_py_types})
    topological_sorter.prepare()
    sorted_py_types: Sequence[PythonType] = ()

    while topological_sorter.is_active():
        node_group = topological_sorter.get_ready()
        sorted_py_types += tuple(sorted(node_group, key=lambda py_type: py_type.name or ""))
        topological_sorter.done(*node_group)

    # Render imports
    content = render_imports(
        tuple({import_ for py_type in sorted_py_types for import_ in py_type.imports}),
    )

    # Render base model class
    content += (
        create_jinja_environment(templates_path=Path(__file__).parent / "templates")
        .get_template("base_model.j2")
        .render()
        + "\n"
    )

    # Render types with a name
    content += "\n".join(py_type.render() for py_type in sorted_py_types if py_type.name is not None)

    models_path = output_dir.joinpath("models.py")
    models_path.write_text(content)

    return models_path



from pydantic import BaseModel, ConfigDict


class Model(BaseModel):
    model_config = ConfigDict(
        frozen=True,  # make instance immutable and hashable
        strict=True,  # strict validation is applied to all fields to avoid coercing values
    )



import argparse
import shutil
import subprocess
from pathlib import Path

from openapi_spec_validator import validate

from spec2sdk.client.generators import generate_client
from spec2sdk.models.generators import generate_models
from spec2sdk.models.imports import Import
from spec2sdk.openapi.parsers import parse_spec
from spec2sdk.openapi.resolver import ResolvingParser


def format_files(path: Path):
    def run_formatter(formatter: str):
        print(f"Running {formatter} on {path}")
        subprocess.run(f"{formatter} {path}", shell=True, check=True)

    # Formatting and linting explanation:
    # 1. Format the code [line will be a 120 character]
    # 2. Add trailing comma [line will be a 121 character]
    # 3. Format the code [line will be a 120 character]
    # 4. Apply linting auto fixes, and then check if the generated code follow the linting rules
    run_formatter("ruff format")
    run_formatter("ruff check --select COM812 --fix")
    run_formatter("ruff format")
    run_formatter("ruff check --fix")


def generate(schema_url: str, output_dir: Path):
    schema = ResolvingParser().parse(schema_url=schema_url)
    validate(schema)
    spec = parse_spec(schema)

    if output_dir.exists():
        shutil.rmtree(str(output_dir))

    output_dir.mkdir(parents=True, exist_ok=True)
    output_dir.joinpath("__init__.py").write_text("")

    models_path = generate_models(spec=spec, output_dir=output_dir)
    generate_client(spec=spec, models_import=Import(name="", package=f".{models_path.stem}"), output_dir=output_dir)

    format_files(output_dir)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--output-dir",
        type=Path,
        help="Path to the output directory where the generated code will be written to",
        required=True,
    )

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--schema-path",
        type=Path,
        help="File path to the OpenAPI specification file in YAML format",
    )
    group.add_argument(
        "--schema-url",
        type=str,
        help="URL of the OpenAPI specification file in YAML format",
    )

    args = parser.parse_args()
    generate(
        schema_url=args.schema_path.absolute().as_uri() if args.schema_path else args.schema_url,
        output_dir=args.output_dir,
    )


if __name__ == "__main__":
    main()
